from .sparrow_message import SparrowData


class InvalidDataPushException(Exception):
    def __init(self, sparrow_data: SparrowData):
        self.sparrow_data = sparrow_data

    def __str__(self):
        return str(self.sparrow_data.get_dict())


class CommandNotFound(Exception):
    def __init__(self, command: str):
        self.command = command

    def __str__(self):
        return "{}는 존재하지 않는 명령어입니다. ".format(self.command)



import logging
from typing import Union

from .core.queue.sparrow_client_queue import SparrowClientQueue
from .core.slack_command import SlackCommandManager
from .core.sparrow_exception import CommandNotFound
from .core.sparrow_message import (
    Color,
    SparrowData,
)

logger = logging.getLogger(__name__)


def send_slack(channel: str, message: Union[str, list, dict], title: str = "", color: str = Color.SECONDARY):
    """
    Args:
        channel: 메시지를 전송할 channel의 이름
        message: 메시지의 본문
        title: 메시지의 타이틀
        color: slack 메시지의 quote 컬러
    """
    total_len = 0
    sparrow_datas = []
    text_buffer = []

    if isinstance(message, list):
        for item in message:
            text = f">`{item}` \n"
            text_buffer.append(text)
    elif isinstance(message, dict):
        for key, value in message:
            text = f">`{key}`: *{value}*` \n"
            text_buffer.append(text)
    else:
        text_buffer.append(message)

    texts = []
    if len(text_buffer) > 1:
        for text in text_buffer:
            text_len = len(text)
            if total_len + text_len > 4000:
                sparrow_data = SparrowData(channel=channel, title_text=title, color=color, body_text=''.join(texts))
                sparrow_datas.append(sparrow_data)
                total_len = 0
                texts = []
            total_len += text_len
            texts.append(text)
    else:
        texts = text_buffer

    sparrow_data = SparrowData(channel=channel, title_text=title, color=color, body_text=''.join(texts))
    sparrow_datas.append(sparrow_data)

    for sparrow_data in sparrow_datas:
        SparrowClientQueue().push(sparrow_data)

    return


def execute_command(sparrow_data: SparrowData):
    """ 사용자가 입력한 명령어가 존재하는 명령어인지 확인 후, 해당 명령어에 대한 response의 결과를 슬랙으로 전송
    """
    cmd = SlackCommandManager.get_command(sparrow_data.command)

    if cmd is None:
        logging.exception("command not found")
        raise CommandNotFound(sparrow_data.command)

    message = cmd.response(
        user_command=sparrow_data.command,
        channel=sparrow_data.channel,
        username=sparrow_data.username
    )
    send_slack(channel=sparrow_data.channel, message=message, color=cmd.color)
    return



import json
import os

from ..queue.redis_queue import RedisQueue
from ..sparrow_exception import InvalidDataPushException
from ..sparrow_message import SparrowData


class Singleton(type):
    __instance = None

    def __call__(cls, *args, **kwargs):
        if cls.__instance is None:
            cls.__instance = super().__call__(*args, **kwargs)
        return cls.__instance


class SparrowClientQueue(RedisQueue, metaclass=Singleton):

    def __init__(self, **kwargs):
        host = os.getenv("SPARROW_REDIS_HOST")
        port = int(os.getenv("SPARROW_REDIS_PORT"))
        db = int(os.getenv("SPARROW_REDIS_DB"))

        super().__init__(host, port, db, **kwargs)

    def push(self, sparrow_data: SparrowData):
        sp_dict = sparrow_data.get_dict()

        if not _is_valid_sparrow_data(sp_dict):
            raise InvalidDataPushException(sparrow_data)

        sp_data = json.dumps(sp_dict)

        if isinstance(sp_data, bytes):
            sp_data = sp_data.decode('utf-8')

        self.enqueue(sparrow_data.channel, sp_data)
        return


def _is_valid_sparrow_data(sp_dict: dict) -> bool:
    """
    redis에서 push 직전, pop 직후의 dict 데이터를 통한 유효성 검증
    """
    if not (isinstance(sp_dict, dict)
            and isinstance(sp_dict.get("username"), str)
            and isinstance(sp_dict.get("command"), str)

            and isinstance(sp_dict.get("channel"), str)
            and isinstance(sp_dict.get("title_text"), str)

            and isinstance(sp_dict.get("color"), str)
            and isinstance(sp_dict.get("body_text"), str)):
        return False
    return True



from abc import (
    ABCMeta,
    abstractmethod,
)
from typing import (
    Optional,
    Union,
)

from .sparrow_message import Color


class SlackCommand(metaclass=ABCMeta):
    """
    Base Class of User's Command.
    """

    @property
    def color(self) -> str:
        return Color.SUCCESS

    @property
    @abstractmethod
    def execution_words(self) -> Union[list, tuple]:
        """ 특정 명령을 실행할 수 있는 명령어를 반환

        Returns:
            이 명령을 실행할 수 있는 단어의 튜플로 부합하는 명령이 들어오면, `response` 함수를 실행
        """

    @abstractmethod
    def response(self, user_command: str, channel: str, username: str) -> str:
        """ 명령이 실행되는 함수

        Returns:
            slack으로 보내야하는 response 메시지.

        """


class SlackCommandManager:
    commands = []

    @classmethod
    def register(cls, command: SlackCommand):
        cls.commands.append(command)

    @classmethod
    def get_command(cls, command: str) -> Optional[SlackCommand]:
        for cmd in cls.commands:
            if command in cmd.execution_words:
                return cmd
        return None



from enum import EnumMeta


class Color(EnumMeta):
    PRIMARY = "#2b7cf6"
    SECONDARY = "#6d757c"
    SUCCESS = "#53a451"
    DANGER = "#cc444a"
    WARNING = "#f7c244"


class SparrowData:
    # inapi python 버전이 낮아 dataclass를 사용하지 못 함
    """
    workspace: str = ""
    username: str = ""
    command: str = ""

    channel: str = ""
    title_text: str = ""

    color: str = Color.SECONDARY
    body_text: str = ""
    """
    def __init__(self, workspace="", username="", command="", channel="", title_text="", color=Color.SECONDARY, body_text=""):
        self.workspace = workspace
        self.username = username
        self.command = command
        self.channel = channel
        self.title_text = title_text
        self.color = color
        self.body_text = body_text


    def get_dict(self) -> dict:
        return self.__dict__



from typing import Optional

import redis


class RedisQueue:
    def __init__(self, host: str, port: int, db: int, **kwargs):
        pool = redis.ConnectionPool(
            host=host, port=port, db=db, max_connections=10, **kwargs
        )
        self._redis = redis.StrictRedis(charset="utf-8", decode_responses=True, connection_pool=pool)

    def get_connection(self) -> redis.StrictRedis:
        return self._redis

    def get_list(self, key) -> list:
        return self._redis.lrange(key, 0, -1)

    def count(self, key, value):
        """ 특정 키 내의 동일한 value의 개수를 반환합니다. """
        return self.get_list(key).count(value)

    def enqueue(self, key, value):
        self._redis.rpush(key, value)

    def dequeue(self, key) -> Optional[str]:
        """ 키에 value가 존재하지 않는 경우 None """
        return self._redis.lpop(key)

    def remove(self, key, value) -> int:
        """ key의 list에서 동일한 value를 모두 지우고 그 개수를 반환 """
        return self._redis.lrem(key, 0, value)

    def _clear_queue(self, key) -> int:
        return self._redis.delete(key)

    def size(self, key) -> int:
        return self._redis.llen(key)

    def get_keys(self) -> list:
        return self._redis.keys()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="sparrow-client", # Replace with your own username
    version="1.0.4",
    author="yeoseot",
    author_email="changyeon@peoplefund.co.kr",
    description="sparrow_client",
    long_description="",
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
    install_requires=[
        'redis==3.3.8',
        'requests==2.22.0',
    ]
)
